@@ -38,7 +38,7 @@
 #include "display.h"
 #include "auth.h"
 
-static const gchar RCSid[]="$Id: server.c,v 1.17 2001/04/30 05:45:31 jirka Exp $";
+static const gchar RCSid[]="$Id: server.c,v 1.20 2001/05/17 08:26:20 jirka Exp $";
 
 #define SERVER_WAIT_ALARM 10
 
@@ -153,47 +153,47 @@ gdm_server_start (GdmDisplay *disp)
     /* Create new cookie */
     gdm_auth_secure_display (d);
     gdm_setenv ("DISPLAY", d->name);
-    
+
     /* Catch USR1 from X server */
     usr1.sa_handler = gdm_server_usr1_handler;
     usr1.sa_flags = SA_RESTART|SA_RESETHAND;
     sigemptyset (&usr1.sa_mask);
-    
+
     if (sigaction (SIGUSR1, &usr1, &old_usr1) < 0) {
-	gdm_error (_("gdm_server_start: Error setting up USR1 signal handler"));
-	return FALSE;
+	    gdm_error (_("gdm_server_start: Error setting up USR1 signal handler"));
+	    return FALSE;
     }
 
     /* Catch CHLD from X server */
     chld.sa_handler = gdm_server_child_handler;
     chld.sa_flags = SA_RESTART|SA_RESETHAND;
     sigemptyset (&chld.sa_mask);
-    
+
     if (sigaction (SIGCHLD, &chld, &old_chld) < 0) {
-	gdm_error (_("gdm_server_start: Error setting up CHLD signal handler"));
-	sigaction (SIGUSR1, &old_usr1, NULL);
-	return FALSE;
+	    gdm_error (_("gdm_server_start: Error setting up CHLD signal handler"));
+	    sigaction (SIGUSR1, &old_usr1, NULL);
+	    return FALSE;
     }
 
     /* Catch ALRM from X server */
     alrm.sa_handler = gdm_server_alarm_handler;
     alrm.sa_flags = SA_RESTART|SA_RESETHAND;
     sigemptyset (&alrm.sa_mask);
-    
+
     if (sigaction (SIGALRM, &alrm, &old_alrm) < 0) {
-	gdm_error (_("gdm_server_start: Error setting up ALRM signal handler"));
-	sigaction (SIGUSR1, &old_usr1, NULL);
-	sigaction (SIGCHLD, &old_chld, NULL);
-	return FALSE;
+	    gdm_error (_("gdm_server_start: Error setting up ALRM signal handler"));
+	    sigaction (SIGUSR1, &old_usr1, NULL);
+	    sigaction (SIGCHLD, &old_chld, NULL);
+	    return FALSE;
     }
-    
+
     /* Set signal mask */
     sigemptyset (&mask);
     sigaddset (&mask, SIGUSR1);
     sigaddset (&mask, SIGCHLD);
     sigaddset (&mask, SIGALRM);
     sigprocmask (SIG_UNBLOCK, &mask, &oldmask);
-    
+
     /* We add a timeout in case the X server fails to start. This
      * might happen because X servers take a while to die, close their
      * sockets etc. If the old X server isn't completely dead, the new
@@ -202,7 +202,7 @@ gdm_server_start (GdmDisplay *disp)
     alarm (SERVER_WAIT_ALARM);
 
     d->servstat = SERVER_DEAD;
-    
+
     /* fork X server process */
     gdm_server_spawn (d);
 
@@ -241,7 +241,8 @@ gdm_server_start (GdmDisplay *disp)
 		    waitpid (d->servpid, NULL, 0);
 	    d->servpid = 0;
     }
-    _exit (DISPLAY_REMANAGE);
+
+    _exit (DISPLAY_XFAILED);
 
 spawn_done:
 
@@ -266,7 +267,7 @@ spawn_done:
 static void 
 gdm_server_spawn (GdmDisplay *d)
 {
-    struct sigaction usr1, dfl_signal;
+    struct sigaction ign_signal, dfl_signal;
     sigset_t mask;
     gchar *srvcmd = NULL;
     gchar **argv = NULL;
@@ -307,18 +308,26 @@ gdm_server_spawn (GdmDisplay *d)
 	}
 
 	
-	/* The X server expects USR1 to be SIG_IGN */
-	usr1.sa_handler = SIG_IGN;
-	usr1.sa_flags = SA_RESTART;
-	sigemptyset (&usr1.sa_mask);
+	/* The X server expects USR1/TTIN/TTOU to be SIG_IGN */
+	ign_signal.sa_handler = SIG_IGN;
+	ign_signal.sa_flags = SA_RESTART;
+	sigemptyset (&ign_signal.sa_mask);
 
-	if (sigaction (SIGUSR1, &usr1, NULL) < 0) {
+	if (sigaction (SIGUSR1, &ign_signal, NULL) < 0) {
 	    gdm_error (_("gdm_server_spawn: Error setting USR1 to SIG_IGN"));
 	    _exit (SERVER_ABORT);
 	}
+	if (sigaction (SIGTTIN, &ign_signal, NULL) < 0) {
+	    gdm_error (_("gdm_server_spawn: Error setting TTIN to SIG_IGN"));
+	    _exit (SERVER_ABORT);
+	}
+	if (sigaction (SIGTTOU, &ign_signal, NULL) < 0) {
+	    gdm_error (_("gdm_server_spawn: Error setting TTOU to SIG_IGN"));
+	    _exit (SERVER_ABORT);
+	}
 
 	/* And HUP and TERM should be at default */
-	dfl_signal.sa_handler = SIG_IGN;
+	dfl_signal.sa_handler = SIG_DFL;
 	dfl_signal.sa_flags = SA_RESTART;
 	sigemptyset (&dfl_signal.sa_mask);
 
@@ -460,6 +469,8 @@ gdm_server_alloc (gint id, const gchar *command)
     d->sessionid = 0;
     d->acctime = 0;
     d->dsp = NULL;
+    d->screenx = 0; /* xinerama offset */
+    d->screeny = 0;
 
     d->last_start_time = 0;
     d->retry_count = 0;
@@ -474,24 +485,24 @@ gdm_server_alloc (gint id, const gchar *command)
 }
 
 /* ignore handlers */
-static gint
+static int
 ignore_xerror_handler (Display *disp, XErrorEvent *evt)
 {
-    return 0;
+	return 0;
 }
 
-static gint
+static int
 ignore_xioerror_handler (Display *disp)
 {
-    return 0;
+	return 0;
 }
 
 void
 gdm_server_whack_clients (GdmDisplay *disp)
 {
 	int i, screen_count;
-	gint (* old_xerror_handler) (Display *, XErrorEvent *);
-	gint (* old_xioerror_handler) (Display *);
+	int (* old_xerror_handler) (Display *, XErrorEvent *);
+	int (* old_xioerror_handler) (Display *);
 
 	if (disp == NULL ||
 	    disp->dsp == NULL)
@@ -521,6 +532,7 @@ gdm_server_whack_clients (GdmDisplay *disp)
 		}
 	}
 
+	XSync (disp->dsp, False);
 	XSetErrorHandler (old_xerror_handler);
 	XSetIOErrorHandler (old_xioerror_handler);
 }
