@@ -21,6 +21,7 @@
 
 #include <config.h>
 #include <gnome.h>
+#include <gdk/gdkx.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <fcntl.h>
@@ -29,6 +30,9 @@
 #include <sys/wait.h>
 #include <strings.h>
 #include <X11/Xlib.h>
+#ifdef HAVE_LIBXINERAMA
+#include <X11/extensions/Xinerama.h>
+#endif
 #include <signal.h>
 #include <pwd.h>
 #include <grp.h>
@@ -44,7 +48,7 @@
 #include "server.h"
 
 
-static const gchar RCSid[]="$Id: slave.c,v 1.39 2001/05/02 10:14:38 jirka Exp $";
+static const gchar RCSid[]="$Id: slave.c,v 1.51 2001/05/21 02:31:49 jirka Exp $";
 
 /* Some per slave globals */
 static GdmDisplay *d;
@@ -53,11 +57,13 @@ static sigset_t mask;
 static gboolean greet = FALSE;
 static FILE *greeter;
 static pid_t last_killed_pid = 0;
-gboolean do_timed_login = FALSE; /* if this is true, login the timed login */
-gboolean do_configurator = FALSE; /* if this is true, login as root
-				   * and start the configurator */
+static gboolean do_timed_login = FALSE; /* if this is true,
+					   login the timed login */
+static gboolean do_configurator = FALSE; /* if this is true, login as root
+					  * and start the configurator */
 
 extern gboolean gdm_first_login;
+extern gboolean gdm_emergency_server;
 
 /* Configuration option variables */
 extern gchar *GdmUser;
@@ -69,6 +75,7 @@ extern gchar *GdmAutomaticLogin;
 extern gchar *GdmConfigurator;
 extern gboolean GdmConfigAvailable;
 extern gboolean GdmSystemMenu;
+extern gint GdmXineramaScreen;
 extern gchar *GdmGreeter;
 extern gchar *GdmDisplayInit;
 extern gchar *GdmPreSession;
@@ -88,6 +95,8 @@ extern gint GdmRetryDelay;
 extern gboolean GdmAllowRoot;
 extern sigset_t sysmask;
 extern gchar *argdelim;
+extern gchar *GdmGlobalFaceDir;
+extern gboolean GdmBrowser;
 
 
 /* Local prototypes */
@@ -189,6 +198,72 @@ setup_automatic_session (GdmDisplay *display, const char *name)
 	gdm_debug ("gdm_slave_start: DisplayInit script finished");
 }
 
+#ifdef HAVE_LIBXINERAMA
+/* Yay thread unsafety */
+static gboolean x_error_occured = FALSE;
+
+/* ignore handlers */
+static int
+ignore_xerror_handler (Display *disp, XErrorEvent *evt)
+{
+	x_error_occured = TRUE;
+	return 0;
+}
+
+static int
+ignore_xioerror_handler (Display *disp)
+{
+	x_error_occured = TRUE;
+	return 0;
+}
+#endif
+
+static void 
+gdm_screen_init (GdmDisplay *display) 
+{
+#ifdef HAVE_LIBXINERAMA
+	int (* old_xerror_handler) (Display *, XErrorEvent *);
+	int (* old_xioerror_handler) (Display *);
+	gboolean have_xinerama = FALSE;
+
+	x_error_occured = FALSE;
+	old_xerror_handler = XSetErrorHandler (ignore_xerror_handler);
+	old_xioerror_handler = XSetIOErrorHandler (ignore_xioerror_handler);
+
+	have_xinerama = XineramaIsActive (display->dsp);
+
+	XSync (display->dsp, False);
+	XSetErrorHandler (old_xerror_handler);
+	XSetIOErrorHandler (old_xioerror_handler);
+
+	if (x_error_occured)
+		have_xinerama = FALSE;
+
+	if (have_xinerama) {
+		int screen_num;
+		XineramaScreenInfo *xscreens =
+			XineramaQueryScreens (display->dsp,
+					      &screen_num);
+
+
+		if (screen_num <= 0)
+			gdm_fail ("Xinerama active, but <= 0 screens?");
+
+		if (screen_num <= GdmXineramaScreen)
+			GdmXineramaScreen = 0;
+
+		display->screenx = xscreens[GdmXineramaScreen].x_org;
+		display->screeny = xscreens[GdmXineramaScreen].y_org;
+
+		XFree (xscreens);
+	} else
+#endif
+	{
+		display->screenx = 0;
+		display->screeny = 0;
+	}
+}
+
 
 static void 
 gdm_slave_run (GdmDisplay *display)
@@ -234,6 +309,10 @@ gdm_slave_run (GdmDisplay *display)
     }
 
     if (d->dsp != NULL) {
+
+	/* checkout xinerama */
+        gdm_screen_init (d);
+
 	if (d->type == TYPE_LOCAL &&
 	    gdm_first_login &&
 	    ! gdm_string_empty (GdmAutomaticLogin) &&
@@ -288,6 +367,84 @@ gdm_slave_whack_greeter (void)
 	sigprocmask (SIG_SETMASK, &omask, NULL);
 }
 
+/* A hack really, this will wait around until the first mapped window
+ * with this class and focus it */
+static void
+focus_first_x_window (const char *class_res_name)
+{
+	pid_t pid;
+	Display *disp;
+	XWindowAttributes attribs = { 0, };
+
+	pid = fork ();
+	if (pid < 0) {
+		gdm_error (_("focus_first_x_window: cannot fork"));
+		return;
+	}
+	/* parent */
+	if (pid > 0) {
+		return;
+	}
+
+	disp = XOpenDisplay (d->name);
+	if (disp == NULL) {
+		gdm_error (_("focus_first_x_window: cannot open display %s"),
+			   d->name);
+		_exit (0);
+	}
+
+	/* set event mask for events on root window */
+	XGetWindowAttributes (disp,
+			      DefaultRootWindow (disp),
+			      &attribs);
+	XSelectInput (disp,
+		      DefaultRootWindow (disp),
+		      attribs.your_event_mask |
+		      SubstructureNotifyMask);
+	
+	for (;;) {
+		XEvent event = { 0, };
+		XClassHint hint = { NULL, NULL };
+
+		XNextEvent (disp, &event);
+
+		if (event.type == MapNotify &&
+		    XGetClassHint (disp,
+				   event.xmap.window,
+				   &hint) &&
+		    hint.res_name != NULL &&
+		    strcmp (hint.res_name, class_res_name) == 0) {
+			XSetInputFocus (disp,
+					event.xmap.window,
+					RevertToPointerRoot,
+					CurrentTime);
+			XSync (disp, True);
+			XCloseDisplay (disp);
+
+			_exit (0);
+		}
+	}
+}
+
+static gboolean
+accept_both_clicks (GtkWidget *w,
+		    GdkEvent *event,
+		    gpointer data)
+{
+	/* HAAAAAAAAAAAAAAAAACK */
+	/* Since the user has not logged in yet and may have left/right
+	 * mouse buttons switched, we just translate every right mouse click
+	 * to a left mouse click */
+	if ((event->type == GDK_BUTTON_PRESS ||
+	     event->type == GDK_2BUTTON_PRESS ||
+	     event->type == GDK_3BUTTON_PRESS ||
+	     event->type == GDK_BUTTON_RELEASE)
+	    && event->button.button == 3)
+		event->button.button = 1;
+
+	return FALSE;
+}      
+
 /* A hack really, this pretends to be a standalone gtk program */
 /* this should only be called once forked and all thingies are closed */
 static void
@@ -299,35 +456,64 @@ run_error_dialog (const char *error)
 	GtkWidget *dialog;
 	GtkWidget *label;
 	GtkWidget *button;
+	pid_t pid;
 
-	gtk_init (&argc, &argv);
+	pid = fork ();
+	/* if we can't fork or we are a child */
+	if (pid <= 0) {
+		gtk_init (&argc, &argv);
 
-	dialog = gtk_dialog_new ();
+		dialog = gtk_dialog_new ();
+		gtk_widget_set_uposition (dialog, d->screenx, d->screeny);
 
-	gtk_signal_connect (GTK_OBJECT (dialog), "destroy",
-			    GTK_SIGNAL_FUNC(gtk_main_quit),
-			    NULL);
+		gtk_signal_connect (GTK_OBJECT (dialog), "destroy",
+				    GTK_SIGNAL_FUNC(gtk_main_quit),
+				    NULL);
 
-	gtk_window_set_title (GTK_WINDOW (dialog), _("Cannot start session"));
+		gtk_window_set_title (GTK_WINDOW (dialog), _("Cannot start session"));
 
-	label = gtk_label_new (error);
+		label = gtk_label_new (error);
 
-	gtk_container_set_border_width
-		(GTK_CONTAINER (GTK_DIALOG (dialog)->vbox), 10);
-	gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->vbox), label,
-			    TRUE, TRUE, 0);
+		gtk_container_set_border_width
+			(GTK_CONTAINER (GTK_DIALOG (dialog)->vbox), 10);
+		gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->vbox), label,
+				    TRUE, TRUE, 0);
 
-	button = gtk_button_new_with_label (_("OK"));
-	gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->action_area),
-			    button, TRUE, TRUE, 0);
+		button = gtk_button_new_with_label (_("OK"));
+		gtk_signal_connect (GTK_OBJECT (button), "event",
+				    GTK_SIGNAL_FUNC (accept_both_clicks),
+				    NULL);
+		gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->action_area),
+				    button, TRUE, TRUE, 0);
 
-	gtk_signal_connect_object (GTK_OBJECT (button), "clicked",
-				   GTK_SIGNAL_FUNC (gtk_widget_destroy), 
-				   GTK_OBJECT (dialog));
+		gtk_signal_connect_object (GTK_OBJECT (button), "clicked",
+					   GTK_SIGNAL_FUNC (gtk_widget_destroy), 
+					   GTK_OBJECT (dialog));
+		GTK_WIDGET_SET_FLAGS (button, GTK_CAN_DEFAULT);
+		gtk_widget_grab_default (button);
 
-	gtk_widget_show_all (dialog);
+		gtk_widget_show_all (dialog);
+		gtk_widget_show_now (dialog);
 
-	gtk_main ();
+		if (dialog->window != NULL) {
+			gdk_error_trap_push ();
+			XSetInputFocus (GDK_DISPLAY (),
+					GDK_WINDOW_XWINDOW (dialog->window),
+					RevertToPointerRoot,
+					CurrentTime);
+			gdk_flush ();
+			gdk_error_trap_pop ();
+		}
+
+		gtk_main ();
+
+		if (pid == 0)
+			_exit (0);
+	}
+
+	if (pid > 0) {
+		waitpid (pid, 0, 0);
+	}
 }
 
 static void
@@ -358,6 +544,7 @@ run_config (GdmDisplay *display, struct passwd *pwent)
 		gdm_setenv ("HOME", pwent->pw_dir);
 		gdm_setenv ("SHELL", pwent->pw_shell);
 		gdm_setenv ("PATH", GdmRootPath);
+		gdm_setenv ("RUNNING_UNDER_GDM", "true");
 
 		for (i = 0; i < sysconf (_SC_OPEN_MAX); i++)
 			close(i);
@@ -369,7 +556,20 @@ run_config (GdmDisplay *display, struct passwd *pwent)
 		open("/dev/null", O_RDWR); /* open stderr - fd 2 */
 
 		/* exec the configurator */
-		argv = g_strsplit (GdmConfigurator, " ", MAX_ARGS);	
+		argv = g_strsplit (GdmConfigurator, " ", MAX_ARGS);
+		if (access (argv[0], X_OK) == 0)
+			execv (argv[0], argv);
+
+		run_error_dialog (_("Could not execute the configuration\n"
+				    "program.  Make sure it's path is set\n"
+				    "correctly in the configuration file.\n"
+				    "I will attempt to start it from the\n"
+				    "default location."));
+
+		argv = g_strsplit
+			(EXPANDED_GDMCONFIGDIR
+			 "/gdmconfig --disable-sound --disable-crash-dialog",
+			 " ", MAX_ARGS);
 		if (access (argv[0], X_OK) == 0)
 			execv (argv[0], argv);
 
@@ -478,7 +678,7 @@ gdm_slave_wait_for_login (void)
 			gdm_verify_cleanup ();
 
 			gdm_slave_greeter_ctl_no_ret (GDM_ENABLE, "");
-			gdm_slave_greeter_ctl_no_ret (GDM_RESET, "");
+			gdm_slave_greeter_ctl_no_ret (GDM_RESETOK, "");
 			continue;
 		}
 
@@ -506,12 +706,125 @@ gdm_slave_wait_for_login (void)
 	}
 }
 
+/* This is VERY evil! */
+static void
+run_pictures (void)
+{
+	char *response;
+	int tempfd;
+	char tempname[256];
+	char buf[1024];
+	size_t bytes;
+	struct passwd *pwent;
+	char *picfile;
+	FILE *fp;
+
+	for (;;) {
+		response = gdm_slave_greeter_ctl (GDM_NEEDPIC, "");
+		if (gdm_string_empty (response)) {
+			g_free (response);
+			return;
+		}
+		
+		/* don't quit, we don't want to further confuse a confused
+		 * greeter, just don't send it pics */
+		if ( ! GdmBrowser) {
+			gdm_slave_greeter_ctl_no_ret (GDM_READPIC, "");
+			continue;
+		}
+
+		pwent = getpwnam (response);
+		if (pwent == NULL) {
+			gdm_slave_greeter_ctl_no_ret (GDM_READPIC, "");
+			continue;
+		}
+
+		picfile = g_strconcat (pwent->pw_dir, "/.gnome/photo", NULL);
+		if (access (picfile, F_OK) != 0) {
+			g_free (picfile);
+			picfile = g_strconcat (GdmGlobalFaceDir, "/",
+					       response, NULL);
+	
+			if (access (picfile, R_OK) == 0) {
+				gdm_slave_greeter_ctl_no_ret (GDM_READPIC,
+							      picfile);
+				g_free (picfile);
+				continue;
+			}
+
+			gdm_slave_greeter_ctl_no_ret (GDM_READPIC, "");
+			g_free (picfile);
+			continue;
+		}
+		g_free (picfile);
+
+		setegid (pwent->pw_gid);
+		seteuid (pwent->pw_uid);
+
+		/* Sanity check on ~user/.gnome/photo */
+		picfile = g_strconcat (pwent->pw_dir, "/.gnome", NULL);
+		if ( ! gdm_file_check ("run_pictures", pwent->pw_uid,
+				       picfile, "photo", TRUE, GdmUserMaxFile,
+				       GdmRelaxPerms)) {
+			g_free (picfile);
+
+			seteuid (0);
+			setegid (GdmGroupId);
+
+			gdm_slave_greeter_ctl_no_ret (GDM_READPIC, "");
+			continue;
+		}
+		g_free (picfile);
+		picfile = g_strconcat (pwent->pw_dir, "/.gnome/photo", NULL);
+
+		fp = fopen (picfile, "r");
+		g_free (picfile);
+		if (fp == NULL) {
+			seteuid (0);
+			setegid (GdmGroupId);
+
+			gdm_slave_greeter_ctl_no_ret (GDM_READPIC, "");
+			continue;
+		}
+
+		strcpy (tempname, "/tmp/gdm-user-picture-XXXXXX");
+		tempfd = mkstemp (tempname);
+
+		if (tempfd < 0) {
+			fclose (fp);
+
+			seteuid (0);
+			setegid (GdmGroupId);
+
+			gdm_slave_greeter_ctl_no_ret (GDM_READPIC, "");
+			continue;
+		}
+
+		fchmod (tempfd, 0644);
+
+		while ((bytes = fread (buf, sizeof (char),
+				       sizeof (buf), fp)) > 0) {
+			write (tempfd, buf, bytes);
+		}
+
+		fclose (fp);
+		close (tempfd);
+
+		gdm_slave_greeter_ctl_no_ret (GDM_READPIC, tempname);
+
+		unlink (tempname);
+
+		seteuid (0);
+		setegid (GdmGroupId);
+	}
+}
 
 static void
 gdm_slave_greeter (void)
 {
     gint pipe1[2], pipe2[2];  
     gchar **argv;
+    struct passwd *pwent;
     
     gdm_debug ("gdm_slave_greeter: Running greeter on %s", d->name);
     
@@ -559,8 +872,25 @@ gdm_slave_greeter (void)
 	gdm_clearenv_no_lang ();
 	gdm_setenv ("XAUTHORITY", d->authfile);
 	gdm_setenv ("DISPLAY", d->name);
-	gdm_setenv ("HOME", "/"); /* Hack */
+
+	gdm_setenv ("LOGNAME", GdmUser);
+	gdm_setenv ("USER", GdmUser);
+	gdm_setenv ("USERNAME", GdmUser);
+	
+	pwent = getpwnam (GdmUser);
+	if (pwent != NULL) {
+		/* Note that usually this doesn't exist */
+		if (g_file_exists (pwent->pw_dir))
+			gdm_setenv ("HOME", pwent->pw_dir);
+		else
+			gdm_setenv ("HOME", "/"); /* Hack */
+		gdm_setenv ("SHELL", pwent->pw_shell);
+	} else {
+		gdm_setenv ("HOME", "/"); /* Hack */
+		gdm_setenv ("SHELL", "/bin/sh");
+	}
 	gdm_setenv ("PATH", GdmDefaultPath);
+	gdm_setenv ("RUNNING_UNDER_GDM", "true");
 
 	/* Note that this is just informative, the slave will not listen to
 	 * the greeter even if it does something it shouldn't on a non-local
@@ -579,8 +909,45 @@ gdm_slave_greeter (void)
 		gdm_unsetenv ("GDM_TIMED_LOGIN_OK");
 	}
 
+	if(gdm_emergency_server) {
+		run_error_dialog (_("No servers were defined in the\n"
+				    "configuration file and xdmcp was\n"
+				    "disabled.  This can only be a\n"
+				    "configuration error.  So I have started\n"
+				    "a single server for you.  You should\n"
+				    "log in and fix the configuration.\n"
+				    "Note that automatic and timed logins\n"
+				    "are disabled now."));
+		gdm_unsetenv ("GDM_TIMED_LOGIN_OK");
+	}
+
+	if (d->failsafe_xserver) {
+		run_error_dialog (_("I could not start the regular X\n"
+				    "server (your graphical environment)\n"
+				    "and so this is a failsafe X server.\n"
+				    "You should log in and properly\n"
+				    "configure the X server."));
+	}
+
 	argv = g_strsplit (GdmGreeter, argdelim, MAX_ARGS);
 	execv (argv[0], argv);
+
+	gdm_error (_("gdm_slave_greeter: Cannot start greeter trying default: %s"),
+		   EXPANDED_BINDIR
+		   "/gdmlogin --disable-sound --disable-crash-dialog");
+
+	gdm_setenv ("GDM_WHACKED_GREETER_CONFIG", "true");
+
+	argv = g_strsplit (EXPANDED_BINDIR
+			   "/gdmlogin --disable-sound --disable-crash-dialog",
+			   argdelim, MAX_ARGS);
+	execv (argv[0], argv);
+
+	run_error_dialog (_("Cannot start the greeter program,\n"
+			    "you will not be able to log in.\n"
+			    "This display will be disabled.\n"
+			    "Try logging in by other means and\n"
+			    "editting the configuration file"));
 	
 	gdm_slave_exit (DISPLAY_ABORT, _("gdm_slave_greeter: Error starting greeter on display %s"), d->name);
 	
@@ -611,6 +978,8 @@ gdm_slave_greeter (void)
 	greeter = fdopen (STDIN_FILENO, "r");
 	
 	gdm_debug ("gdm_slave_greeter: Greeter on pid %d", d->greetpid);
+
+	run_pictures (); /* Append pictures to greeter if browsing is on */
 	break;
     }
 }
@@ -781,10 +1150,46 @@ find_a_session (void)
 	return session;
 }
 
+static char *
+find_prog (const char *name, const char *args, char **retpath)
+{
+	char *ret;
+	char *path;
+	int i;
+	char *try[] = {
+		"/usr/bin/X11/",
+		"/usr/X11R6/bin/",
+		"/usr/bin/",
+		"/usr/local/bin/",
+		EXPANDED_BINDIR "/",
+		NULL
+	};
+
+	path = gnome_is_program_in_path (name);
+	if (path != NULL &&
+	    access (path, X_OK) == 0) {
+		ret = g_strdup_printf ("%s %s", path, args);
+		*retpath = path;
+		return ret;
+	}
+	g_free (path);
+	for (i = 0; try[i] != NULL; i++) {
+		path = g_strconcat (try[i], name, NULL);
+		if (access (path, X_OK) == 0) {
+			ret = g_strdup_printf ("%s %s", path, args);
+			*retpath = path;
+			return ret;
+		}
+		g_free (path);
+	}
+	*retpath = NULL;
+	return NULL;
+}
+
 static void
 gdm_slave_session_start (void)
 {
-    char *cfgdir, *sesspath;
+    char *cfgdir, *sesspath, *sessexec;
     struct stat statbuf;
     struct passwd *pwent;
     char *save_session = NULL, *session = NULL, *language = NULL, *usrsess, *usrlang;
@@ -888,7 +1293,7 @@ gdm_slave_session_start (void)
 		    savelang = TRUE;
 	    g_free (ret);
 
-	    if (strcmp (session, "Gnome Chooser") == 0) {
+	    if (strcmp (session, GDM_SESSION_GNOME_CHOOSER) == 0) {
 		    char *sessions = gdm_get_sessions (pwent);
 		    ret = gdm_slave_greeter_ctl (GDM_GNOMESESS, sessions);
 		    g_free (sessions);
@@ -1050,7 +1455,7 @@ gdm_slave_session_start (void)
 	/* If "Gnome Chooser" is still set as a session,
 	 * just change that to "Gnome", since "Gnome Chooser" is a
 	 * fake */
-	if (strcmp (session, "Gnome Chooser") == 0) {
+	if (strcmp (session, GDM_SESSION_GNOME_CHOOSER) == 0) {
 		g_free (session);
 		session = g_strdup ("Gnome");
 	}
@@ -1067,45 +1472,55 @@ gdm_slave_session_start (void)
 	sigprocmask (SIG_SETMASK, &sysmask, NULL);
 	
 	sesspath = NULL;
+	sessexec = NULL;
 
 	if (strcmp (session, GDM_SESSION_FAILSAFE_GNOME) == 0) {
-		sesspath = gnome_is_program_in_path ("gnome-session");
+		sesspath = find_prog ("gnome-session",
+				      "--choose-session=Default",
+				      &sessexec);
 		if (sesspath == NULL) {
-			if (g_file_exists ("/usr/bin/X11/gnome-session")) {
-				sesspath = g_strdup ("/usr/bin/X11/gnome-session");
-			} else if (g_file_exists ("/usr/X11R6/bin/gnome-session")) {
-				sesspath = g_strdup ("/usr/X11R6/bin/gnome-session");
-			} else if (g_file_exists ("/usr/bin/gnome-session")) {
-				sesspath = g_strdup ("/usr/bin/gnome-session");
-			} else if (g_file_exists ("/usr/local/bin/gnome-session")) {
-				sesspath = g_strdup ("/usr/local/bin/gnome-session");
-			} else {
-				/* yaikes */
-				gdm_error (_("gdm_slave_session_start: gnome-session not found for a failsafe gnome session, trying xterm"));
-				g_free (session);
-				session = g_strdup (GDM_SESSION_FAILSAFE_XTERM);
-			}
+			/* yaikes */
+			gdm_error (_("gdm_slave_session_start: gnome-session not found for a failsafe gnome session, trying xterm"));
+			g_free (session);
+			session = g_strdup (GDM_SESSION_FAILSAFE_XTERM);
+			run_error_dialog
+				(_("Could not find the GNOME installation,\n"
+				   "will try running the \"Failsafe xterm\"\n"
+				   "session."));
+		} else {
+			run_error_dialog
+				(_("This is the Failsafe Gnome session.\n"
+				   "You will be logged into the 'Default'\n"
+				   "session of Gnome with no startup scripts\n"
+				   "run.  This is only to fix problems in\n"
+				   "your installation."));
 		}
 	}
 
 	/* an if and not an else, we could have done a fall-through
 	 * to here in the above code if we can't find gnome-session */
 	if (strcmp (session, GDM_SESSION_FAILSAFE_XTERM) == 0) {
-		sesspath = gnome_is_program_in_path ("xterm");
+		char *params = g_strdup_printf ("-geometry 80x24+%d+%d",
+						d->screenx, d->screeny);
+		sesspath = find_prog ("xterm",
+				      params,
+				      &sessexec);
+		g_free (params);
 		if (sesspath == NULL) {
-			if (g_file_exists ("/usr/bin/X11/xterm")) {
-				sesspath = g_strdup ("/usr/bin/X11/xterm");
-			} else if (g_file_exists ("/usr/X11R6/bin/xterm")) {
-				sesspath = g_strdup ("/usr/X11R6/bin/xterm");
-			} else if (g_file_exists ("/usr/bin/xterm")) {
-				sesspath = g_strdup ("/usr/bin/xterm");
-			} else {
-				run_error_dialog
-					(_("Cannot find \"xterm\" to start "
-					   "a failsafe session."));
-				/* nyah nyah nyah nyah nyah */
-				_exit (0);
-			}
+			run_error_dialog
+				(_("Cannot find \"xterm\" to start "
+				   "a failsafe session."));
+			/* nyah nyah nyah nyah nyah */
+			_exit (0);
+		} else {
+			run_error_dialog
+				(_("This is the Failsafe xterm session.\n"
+				   "You will be logged into a terminal\n"
+				   "console so that you may fix your system\n"
+				   "if you cannot log in any other way.\n"
+				   "To exit the terminal emulator, type\n"
+				   "'exit' and an enter into the window."));
+			focus_first_x_window ("xterm");
 		}
 	} 
 
@@ -1132,8 +1547,8 @@ gdm_slave_session_start (void)
 		gdm_error (_("gdm_slave_session_start: User not allowed to log in"));
 		run_error_dialog (_("The system administrator has\n"
 				    "disabled your account."));
-	} else if (access (sesspath, X_OK|R_OK) != 0) {
-		gdm_error (_("gdm_slave_session_start: Could not find session `%s'"), sesspath);
+	} else if (access (sessexec != NULL ? sessexec : sesspath, X_OK) != 0) {
+		gdm_error (_("gdm_slave_session_start: Could not find/run session `%s'"), sesspath);
 		/* if we can't read and exec the session, then make a nice
 		 * error dialog */
 		run_error_dialog
@@ -1316,7 +1731,7 @@ gdm_slave_xioerror_handler (Display *disp)
 
     gdm_server_stop (d);
     gdm_verify_cleanup ();
-    _exit (DISPLAY_REMANAGE);
+    _exit (DISPLAY_XFAILED);
 }
 
 char * 
